# test_chunk_node.py
import unittest
from zrag.chunk_node import get_chunk_splitter, Node
from zrag.doc_loader import Document


class TestChunkNode(unittest.TestCase):
    def test_get_chunk_splitter(self):
        splitter = get_chunk_splitter("token", chunk_size=100)
        self.assertIsNotNone(splitter)
        splitter = get_chunk_splitter("overlap", chunk_size=100, overlap=10)
        self.assertIsNotNone(splitter)
        splitter = get_chunk_splitter("paragraph", chunk_size=100)
        self.assertIsNotNone(splitter)

    def test_token_chunk_splitter(self):
        splitter = get_chunk_splitter("token", chunk_size=10)
        doc = Document("1", {"page_label": "1"}, "This is a test document.")
        chunks = splitter.split_document(doc)
        self.assertEqual(len(chunks), 2)

    def test_overlap_chunk_splitter(self):
        splitter = get_chunk_splitter("overlap", chunk_size=20, overlap=5)
        doc = Document("1", {"page_label": "1"}, "This is a test. Another test.")
        chunks = splitter.split_document(doc)
        self.assertEqual(len(chunks), 2)

    def test_paragraph_chunk_splitter(self):
        splitter = get_chunk_splitter("paragraph", chunk_size=20)
        doc = Document("1", {"page_label": "1"}, "This is a test.\n\nAnother test.")
        chunks = splitter.split_document(doc)
        self.assertEqual(len(chunks), 2)

    def test_node_creation(self):
        node = Node("Test text", {"metadata": "value"})
        self.assertEqual(node.text, "Test text")
        self.assertEqual(node.metadata, {"metadata": "value"})


from pathlib import Path
from typing import List, Dict, Optional
import spacy
import fitz
import markdown
import re
import nltk
from zrag.doc_loader import Document


class Node:
    """Represents a single node containing a chunk of text and metadata."""

    def __init__(self, text: str, metadata: Dict):
        """
        Initializes a Node object.

        Args:
            text (str): The text content of the node.
            metadata (Dict): A dictionary containing metadata about the node.
        """
        self.text = text.strip()  # Remove leading/trailing spaces
        self.metadata = metadata

    def __repr__(self):
        return f"Node(text='{self.text[:20]}...', metadata={self.metadata})"


class ChunkSplitter:
    """A base class for different chunking strategies."""

    def split_document(self, document: Document) -> List[Node]:
        """Splits a single document into chunks (Nodes)."""
        raise NotImplementedError("Subclasses should implement this method.")

    def split(self, documents: List[Document]) -> List[Node]:
        """Splits a list of documents into chunks, returning a list of nodes."""
        nodes = []
        for document in documents:
            nodes.extend(self.split_document(document))
        return nodes


class TokenChunkSplitter(ChunkSplitter):
    """Splits documents into chunks of text based on tokens (words)."""

    def __init__(self, chunk_size: int = 256):
        """
        Initializes the TokenChunkSplitter object.

        Args:
            chunk_size (int): The desired size of each chunk (in tokens).
        """
        self.chunk_size = chunk_size
        self.nlp = spacy.load("en_core_web_sm")  # Load a small English model

    def split_document(self, document: Document) -> List[Node]:
        """Splits a document into chunks of text based on tokens."""
        text = document.text
        doc = self.nlp(text)
        tokens = [token.text for token in doc]

        nodes = []
        current_chunk = []
        current_chunk_start_index = 0

        for i, token in enumerate(tokens):
            # If adding the token would exceed the chunk size, start a new chunk
            if len(current_chunk) >= self.chunk_size:
                chunk_text = ' '.join(current_chunk)
                chunk_metadata = {
                    'document_id': document.document_id,
                    'page_label': document.metadata.get('page_label'),
                    'start_index': current_chunk_start_index,
                    'end_index': current_chunk_start_index + len(chunk_text)
                }
                nodes.append(Node(chunk_text, chunk_metadata))
                current_chunk = [token]  # Start a new chunk 
                current_chunk_start_index = doc[i].idx  # Get the character index
            else:
                current_chunk.append(token)

        # Append the last chunk if it's not empty
        if current_chunk:
            chunk_text = ' '.join(current_chunk)
            chunk_metadata = {
                'document_id': document.document_id,
                'page_label': document.metadata.get('page_label'),
                'start_index': current_chunk_start_index,
                'end_index': current_chunk_start_index + len(chunk_text)
            }
            nodes.append(Node(chunk_text, chunk_metadata))

        return nodes


class SentenceChunkSplitterWithOverlap(ChunkSplitter):
    """Splits documents into chunks based on sentences with overlap."""

    def __init__(self, chunk_size: int = 1024, overlap: int = 128):
        """
        Initializes the SentenceChunkSplitterWithOverlap object.

        Args:
            chunk_size (int): The desired size of each chunk (in characters).
            overlap (int): The number of characters to overlap.
        """
        self.chunk_size = chunk_size
        self.overlap = overlap
        nltk.download('punkt', quiet=True)  # Download Punkt sentence tokenizer
        self.tokenizer = nltk.tokenize.PunktSentenceTokenizer()

    def split_document(self, document: Document) -> List[Node]:
        """Splits a document into chunks based on sentences with overlap."""
        text = document.text
        sentences = self.tokenizer.tokenize(text)

        nodes = []
        current_chunk = ""
        current_chunk_start_index = 0

        for sentence in sentences:
            sentence_len = len(sentence)

            # If adding the sentence would exceed the chunk size, start a new chunk
            if len(current_chunk) + sentence_len > self.chunk_size:
                chunk_metadata = {
                    'document_id': document.document_id,
                    'page_label': document.metadata.get('page_label'),
                    'start_index': current_chunk_start_index,
                    'end_index': current_chunk_start_index + len(current_chunk)
                }
                nodes.append(Node(current_chunk.strip(), chunk_metadata))
                
                # Overlap logic
                overlap_start = max(0, len(current_chunk) - self.overlap)
                current_chunk = current_chunk[overlap_start:] + " " + sentence
                current_chunk_start_index += len(current_chunk) - sentence_len - 1 

            else:
                current_chunk += " " + sentence

        # Append the last chunk if it's not empty
        if current_chunk:
            chunk_metadata = {
                'document_id': document.document_id,
                'page_label': document.metadata.get('page_label'),
                'start_index': current_chunk_start_index,
                'end_index': current_chunk_start_index + len(current_chunk)
            }
            nodes.append(Node(current_chunk.strip(), chunk_metadata))

        return nodes


class ParagraphChunkSplitter(ChunkSplitter):
    """Splits documents into chunks of text based on paragraphs."""

    def __init__(self, chunk_size: int = 2048):
        """
        Initializes the ParagraphChunkSplitter object.

        Args:
            chunk_size (int): The desired size of each chunk (in characters).
        """
        self.chunk_size = chunk_size

    def split_document(self, document: Document) -> List[Node]:
        """Splits a document into chunks of text based on paragraphs."""
        text = document.text
        paragraphs = text.split('\n\n')  # Split by double newline

        nodes = []
        current_chunk = ""
        current_chunk_start_index = 0

        for paragraph in paragraphs:
            paragraph_len = len(paragraph)

            # If adding the paragraph would exceed the chunk size, start a new chunk
            if len(current_chunk) + paragraph_len > self.chunk_size:
                chunk_metadata = {
                    'document_id': document.document_id,
                    'page_label': document.metadata.get('page_label'),
                    'start_index': current_chunk_start_index,
                    'end_index': current_chunk_start_index + len(current_chunk)
                }
                nodes.append(Node(current_chunk.strip(), chunk_metadata))
                current_chunk = paragraph  # Start a new chunk 
                current_chunk_start_index += len(current_chunk) + 2  # Add 2 for the double newline
            else:
                current_chunk += paragraph + "\n\n" 

        # Append the last chunk if it's not empty
        if current_chunk:
            chunk_metadata = {
                'document_id': document.document_id,
                'page_label': document.metadata.get('page_label'),
                'start_index': current_chunk_start_index,
                'end_index': current_chunk_start_index + len(current_chunk)
            }
            nodes.append(Node(current_chunk.strip(), chunk_metadata))

        return nodes


def get_chunk_splitter(strategy: str, **kwargs) -> ChunkSplitter:
    """Factory function to select the desired chunking strategy."""
    if strategy == 'token':
        return TokenChunkSplitter(**kwargs)
    elif strategy == 'overlap':
        return SentenceChunkSplitterWithOverlap(**kwargs)
    elif strategy == 'paragraph':
        return ParagraphChunkSplitter(**kwargs)
    else:
        raise ValueError(f"Unknown chunking strategy: {strategy}")


from .doc_loader import DocumentLoader, Document
from .chunk_node import get_chunk_splitter, Node, ChunkSplitter, TokenChunkSplitter, SentenceChunkSplitterWithOverlap, ParagraphChunkSplitter
from .embeddings import Embeddings
from .vector_store import VectorStore
from .llm import LLM
from .prompt_manager import PromptManager
from .rag_pipeline import RAGPipeline
from .data_generation import DataGenerator

__all__ = [
    "DocumentLoader", "Document", 
    "get_chunk_splitter", "Node", "ChunkSplitter", "TokenChunkSplitter", "SentenceChunkSplitterWithOverlap", "ParagraphChunkSplitter",
    "Embeddings",
    "VectorStore",
    "LLM",
    "PromptManager",
    "RAGPipeline",
    "DataGenerator"
]

__version__ = "0.1.2"


import unittest
import numpy as np
import os
from unittest.mock import patch, MagicMock

from zrag.vector_store import VectorStore
from zrag.chunk_node import Node


class TestVectorStore(unittest.TestCase):
    """Comprehensive tests for the VectorStore class."""

    @patch("zrag.vector_store.faiss")
    @patch("zrag.vector_store.chromadb")
    def setUp(self, mock_chroma, mock_faiss):
        """Sets up mocks for FAISS and ChromaDB."""
        self.mock_faiss = mock_faiss
        self.mock_chroma = mock_chroma

    def test_initialization_faiss(self):
        """Tests initialization with FAISS."""
        vs = VectorStore(vector_store_type="faiss")
        self.assertEqual(vs.vector_store_type, "faiss")
        self.assertIsNone(vs.index)

    def test_initialization_chroma(self):
        """Tests initialization with ChromaDB."""
        vs = VectorStore(vector_store_type="chroma")
        self.assertEqual(vs.vector_store_type, "chroma")
        self.assertIsNotNone(vs.client)
        self.assertIsNotNone(vs.collection)

    def test_initialization_invalid(self):
        """Tests initialization with an invalid vector store type."""
        with self.assertRaises(ValueError):
            VectorStore(vector_store_type="invalid")

    def test_index_faiss(self):
        """Tests indexing with FAISS."""
        vs = VectorStore(vector_store_type="faiss")
        chunks = [
            Node("text1", embedding=np.array([0.1, 0.2])),
            Node("text2", embedding=np.array([0.3, 0.4])),
        ]
        vs.index(chunks)
        self.mock_faiss.IndexFlatL2.assert_called_once_with(2)  # Check dimension
        self.mock_faiss.write_index.assert_called_once()

    def test_index_chroma(self):
        """Tests indexing with ChromaDB."""
        vs = VectorStore(vector_store_type="chroma")
        chunks = [
            Node("text1", metadata={"doc_id": "1"}, embedding=np.array([0.1, 0.2])),
            Node("text2", metadata={"doc_id": "2"}, embedding=np.array([0.3, 0.4])),
        ]
        vs.index(chunks)
        vs.collection.add.assert_called_once_with(
            ids=["1", "2"],  # Assuming doc_id is used as node_id
            embeddings=[[0.1, 0.2], [0.3, 0.4]],
            metadatas=[{"doc_id": "1"}, {"doc_id": "2"}],
            documents=["text1", "text2"],
        )

    def test_search_faiss(self):
        """Tests searching with FAISS."""
        vs = VectorStore(vector_store_type="faiss")
        vs.index = MagicMock()
        vs.index.search.return_value = (None, np.array([[0, 1]]))
        query_embedding = np.array([0.1, 0.2])
        results = vs.search(query_embedding)
        vs.index.search.assert_called_once_with(
            np.array([0.1, 0.2]).reshape(1, -1).astype("float32"), 5
        )  # Check default top_k
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]["faiss_index"], 0)
        self.assertEqual(results[1]["faiss_index"], 1)

    def test_search_chroma(self):
        """Tests searching with ChromaDB."""
        vs = VectorStore(vector_store_type="chroma")
        vs.collection.query.return_value = {
            "ids": [["1", "2"]],
            "metadatas": [[{"key": "value1"}, {"key": "value2"}]],
            "distances": [[0.1, 0.2]],
            "documents": [["doc1", "doc2"]],
        }
        query_embedding = np.array([0.1, 0.2])
        results = vs.search(query_embedding)
        vs.collection.query.assert_called_once_with(
            query_embeddings=[[0.1, 0.2]], n_results=5, include=["metadatas", "distances", "documents"]
        )
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0]["node_id"], "1")
        self.assertEqual(results[0]["metadata"], {"key": "value1"})
        self.assertEqual(results[0]["score"], 0.9)  # 1 - distance
        self.assertEqual(results[0]["document"], "doc1")
        # Add assertions for the second result

    def test_save_faiss(self):
        """Tests saving the FAISS index."""
        vs = VectorStore(vector_store_type="faiss")
        vs.index = MagicMock()
        vs.save()
        self.mock_faiss.write_index.assert_called_once_with(vs.index, vs.index_path)

    @patch("os.path.exists", return_value=True)
    def test_load_faiss(self, mock_exists):
        """Tests loading the FAISS index when the file exists."""
        vs = VectorStore(vector_store_type="faiss")
        vs.load()
        self.mock_faiss.read_index.assert_called_once_with(vs.index_path)

    @patch("os.path.exists", return_value=False)
    def test_load_faiss_no_file(self, mock_exists):
        """Tests loading the FAISS index when the file doesn't exist."""
        vs = VectorStore(vector_store_type="faiss")
        vs.load()
        self.mock_faiss.read_index.assert_not_called()


if __name__ == "__main__":
    unittest.main()


import os
import faiss
import logging
import chromadb
import numpy as np
from typing import List, Dict, Any
from zrag.chunk_node import Node  # Updated import

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class VectorStore:
    """Manages indexing and retrieval from vector databases."""

    def __init__(
        self, 
        vector_store_type: str = "faiss", 
        index_path: str = "faiss_index.bin", 
        chroma_persist_dir: str = "chromadb", 
        embedding_dim: int = None
    ):
        """Initializes the VectorStore with either FAISS or ChromaDB."""

        self.vector_store_type = vector_store_type
        self.index_path = index_path
        self.embedding_dim = embedding_dim

        if self.vector_store_type == "faiss":
            self.index = None  # Initialized later 
            logger.info("Initialized FAISS index.")
        elif self.vector_store_type == "chroma":
            self.client = chromadb.PersistentClient(path=chroma_persist_dir)
            self.collection = self.client.get_or_create_collection("documents")
            logger.info("Initialized ChromaDB client and collection.")
        else:
            raise ValueError("Invalid vector_store_type. Choose 'faiss' or 'chroma'.")

    def index(self, chunks: List[Node]):
        """Indexes chunks of text."""
        if self.vector_store_type == "faiss":
            embeddings = np.array([chunk.embedding for chunk in chunks]).astype('float32')
            if self.index is None or self.index.d != embeddings.shape[1]:  # Check if index needs to be created or reset
                self.index = faiss.IndexFlatL2(embeddings.shape[1])
                logger.info("Created/reset FAISS index with correct dimension.")
            self.index.add(embeddings)
            faiss.write_index(self.index, self.index_path)
            logger.info(f"Indexed {len(chunks)} chunks in FAISS and saved to {self.index_path}.")
        elif self.vector_store_type == "chroma":
            ids = [chunk.metadata.get('node_id', str(i)) for i, chunk in enumerate(chunks)]
            embeddings = [chunk.embedding.tolist() for chunk in chunks]  # ChromaDB expects lists
            metadatas = [chunk.metadata for chunk in chunks]  # Pass metadata to ChromaDB
            documents = [chunk.text for chunk in chunks]
            self.collection.add(
                ids=ids,
                embeddings=embeddings,
                metadatas=metadatas,
                documents=documents
            )
            logger.info(f"Indexed {len(chunks)} chunks in ChromaDB.")

    def search(self, query_embedding: np.ndarray, top_k: int = 5) -> List[Dict[str, Any]]:
        """Searches the vector store for similar chunks."""
        if self.vector_store_type == "faiss":
            query_embedding = query_embedding.reshape(1, -1).astype('float32')
            _, indices = self.index.search(query_embedding, top_k)
            results = [{'faiss_index': int(idx)} for idx in indices[0]]
            logger.info(f"FAISS search returned {len(results)} results.")
        elif self.vector_store_type == "chroma":
            results = self.collection.query(
                query_embeddings=[query_embedding.tolist()],
                n_results=top_k,
                include=["metadatas", "distances", "documents"]
            )
            results = [
                {
                    "node_id": result_id,
                    "metadata": metadata,
                    "score": 1 - distance,
                    "document": doc
                }
                for result_id, metadata, distance, doc in zip(
                    results['ids'][0], results['metadatas'][0], results['distances'][0], results['documents'][0]
                )
            ]
            logger.info(f"ChromaDB search returned {len(results)} results.")
        return results

    def save(self):
        """Saves the Faiss index if it exists"""
        if self.vector_store_type == "faiss" and self.index: 
          faiss.write_index(self.index, self.index_path)
          logger.info(f"Saved FAISS index to {self.index_path}")


    def load(self):
        """Loads the Faiss index if it exists"""
        if self.vector_store_type == "faiss":
            if os.path.exists(self.index_path):
                self.index = faiss.read_index(self.index_path)
                logger.info(f"Loaded FAISS index from {self.index_path}")
            else:
                logger.warning(f"Index not found at {self.index_path}")


import logging
from typing import List, Dict, Any, Optional, Union, Generator
from zrag.doc_loader import DocumentLoader
from zrag.chunk_node import get_chunk_splitter, ChunkSplitter, Node 
from zrag.embeddings import Embeddings
from zrag.vector_store import VectorStore
from zrag.llm import LLM
from zrag.prompt_manager import PromptManager

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class RAGPipeline: 
    """
    Implements a Retrieval Augmented Generation (RAG) pipeline.
    """
    def __init__(
        self,
        file_loader: DocumentLoader,
        chunk_splitter: ChunkSplitter,
        embeddings: Embeddings,
        vector_store: VectorStore,
        llm: LLM,
        prompt_manager: PromptManager,
        default_prompt_template: str = "rag_simple",
    ):
        """
        Initializes the RAGPipeline with necessary components.
        """
        self.file_loader = file_loader
        self.chunk_splitter = chunk_splitter
        self.embeddings = embeddings
        self.vector_store = vector_store
        self.llm = llm
        self.prompt_manager = prompt_manager
        self.default_prompt_template = default_prompt_template

    def load_and_index(
        self,
        directory_path: str, 
        recursive: bool = False,
        ext: Optional[List[str]] = None,
        exc: Optional[List[str]] = None,
        filenames: Optional[List[str]] = None,
        preprocess_fn: Optional[Any] = None,
        max_workers: int = None,
    ):
        """Loads, chunks, embeds, and indexes documents from a directory."""
        documents = self.file_loader.load(
            recursive=recursive, 
            ext=ext, 
            exc=exc, 
            filenames=filenames, 
            max_workers=max_workers, 
            preprocess_fn=preprocess_fn
        )
        chunks = self.chunk_splitter.split(documents)
        self.embeddings.embed_nodes(chunks)
        self.vector_store.index(chunks)
        logger.info(f"Loaded, chunked, embedded, and indexed {len(chunks)} chunks from {directory_path}")

    def run(
        self,
        query: str,
        top_k: int = 5,
        prompt_template: Optional[str] = None,
        **llm_kwargs 
    ) -> Union[str, Generator]: 

        """
        Processes a user query and generates a response using the RAG pipeline.
        """
        query_embedding = self.embeddings.embed([query])[0]  
        results = self.vector_store.search(query_embedding, top_k=top_k)  

        # Use prompt_template if provided, otherwise use default
        if not prompt_template:
            prompt_template = self.default_prompt_template

        # Create prompt 
        prompt = self.prompt_manager.create_prompt(
            template_name=prompt_template, query=query, context=results
        )

        if not prompt: 
            logger.error("Prompt creation failed. Returning empty string.")
            return ""  

        # Generate response 
        response = self.llm.generate(prompt, **llm_kwargs)

        return response

    def save_index(self):
        """Saves the vector store index."""
        self.vector_store.save()

    def load_index(self):
        """Loads the vector store index."""
        self.vector_store.load()


import logging
from typing import Dict, Optional, List, Any
from jinja2 import Template

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class PromptManager:
    """
    Manages prompt templates for generating prompts for LLMs.
    """
    def __init__(self, default_template: Optional[str] = None):  
        """Initializes the PromptManager with predefined templates."""
        self.templates = {}
        self._load_predefined_templates()
        if default_template:  
            if default_template not in self.templates:
                raise ValueError(f"Default template '{default_template}' not found in predefined templates.")
            self.default_template_name = default_template
            logger.info(f"Default template set to: {default_template}")
        else:
            self.default_template_name = None

    def _load_predefined_templates(self):
        """Loads predefined prompt templates."""
        self.templates["rag_cot"] = (
            "Context:\n{{ context }}\n\nQuestion: {{ query }}\nLet's think step by step:\nAnswer:"
        )
        self.templates["rag_simple"] = (
            "Context:\n{{ context }}\n\nQuestion: {{ query }}\nAnswer:"
        )
        self.templates["dataset_instruction"] = (
            "Instruction: {{ instruction }}\nInput: {{ input_data }}\nOutput:"
        )
        self.templates["dataset_reasoning"] = "Problem: {{ problem_statement }}\nSolution:"

    def add_template(self, template_name: str, template_str: str):
        """Adds a new prompt template."""
        self.templates[template_name] = template_str 

    def remove_template(self, template_name: str):
        """Removes a prompt template."""
        try:
            del self.templates[template_name]
        except KeyError:  
            logger.warning(f"Template '{template_name}' not found.") 

    def create_prompt(self, template_name: Optional[str] = None, **kwargs) -> str:
        """
        Creates a prompt using a specified template or the default template.
        """

        if template_name is None:
            if self.default_template_name is None:
                logger.error("No template name provided and no default template set.")
                return ""
            template_name = self.default_template_name  

        template_str = self.templates.get(template_name)
        if template_str is None:
            logger.error(f"Template '{template_name}' not found.")
            return ""

        context = kwargs.get('context', [])
        if context:
            if all(isinstance(item, str) for item in context): 
                kwargs['context'] = "\n".join(context)
            elif all(isinstance(item, dict) for item in context) and all('document' in item for item in context):
                # Handle ChromaDB results that include metadata (dictionaries)
                kwargs['context'] = "\n".join([item['document'] for item in context]) 
            else:
                kwargs['context'] = "\n".join([item.text for item in context])
        else:  # If context is not provided or is empty, modify the template to exclude it
            template_str = template_str.replace("Context:\n{{ context }}\n\n", "")

        try:  
            template = Template(template_str)  
            prompt = template.render(**kwargs)
            return prompt
        except Exception as e:
            logger.error(f"Error rendering template: {e}")
            return ""

    def list_templates(self) -> List[str]:
        """Returns a list of available template names."""
        return list(self.templates.keys())


import logging
import torch
from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, AutoModelForCausalLM
from typing import List, Dict, Optional, Union, Iterator, Generator

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class LLM:
    """
    Wrapper for interacting with large language models (LLMs) from Hugging Face.
    """
    def __init__(self, model_name: str, use_gpu: bool = True, task: str = "text-generation", **kwargs):
        """
        Initializes the LLM with the specified model name and task.
        """
        self.model_name = model_name
        self.task = task.lower()  # Ensure task is always lowercase
        self.device = 0 if torch.cuda.is_available() and use_gpu else -1
        self.kwargs = kwargs  # Store additional keyword arguments

        self.tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token

        try:
            try:
                self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name, trust_remote_code=True).to(self.device)
                self.is_encoder_decoder = True
            except ValueError:  
                self.model = AutoModelForCausalLM.from_pretrained(model_name, trust_remote_code=True).to(self.device)
                self.is_encoder_decoder = False
        except Exception as e:  
            logger.error(f"Error loading model: {e}. Check if model name/path is correct.")
            raise

        self.model.eval()

    def _generate_text(self, prompts: Union[str, List[str]], **generation_kwargs) -> Union[str, List[str]]:
        """
        Generates text for a single prompt or list of prompts.
        """

        if isinstance(prompts, str):
            input_ = self.tokenizer(prompts, return_tensors="pt").to(self.device)
            output = self.model.generate(**input_, **generation_kwargs)
            return self.tokenizer.decode(output[0], skip_special_tokens=True)

        elif isinstance(prompts, list):
            outputs = []
            for prompt in prompts:  # Process prompts individually
                input_ = self.tokenizer(prompt, return_tensors="pt").to(self.device)
                output = self.model.generate(**input_, **generation_kwargs)
                outputs.append(self.tokenizer.decode(output[0], skip_special_tokens=True))  
            return outputs
        else:
            raise TypeError("prompts must be a string or a list of strings.")

    def generate(
        self, 
        prompts: Union[str, List[str]], 
        stream_output: bool = False, 
        **generation_kwargs
    ) -> Union[str, List[str], Generator]:
        """
        Generates text for given prompts, with optional streaming output.
        """

        generation_kwargs = {**self.kwargs, **generation_kwargs}  # Combine default and specific kwargs

        # Set defaults based on encoder-decoder model 
        generation_kwargs["max_new_tokens"] = generation_kwargs.get("max_new_tokens", 256 if self.is_encoder_decoder else 512)
        generation_kwargs["num_beams"] = generation_kwargs.get("num_beams", 1 if self.is_encoder_decoder else None)

        # Streaming support
        if stream_output:
            if isinstance(prompts, list):
                return (self._generate_text(prompt, **generation_kwargs) for prompt in prompts) # Streaming
            elif isinstance(prompts, str):
                return self._generate_text(prompts, **generation_kwargs)
            else:
                raise TypeError("prompts must be a string or list of strings.")
        else:
            return self._generate_text(prompts, **generation_kwargs)


import torch
import logging
import numpy as np
from typing import List
from transformers import AutoTokenizer, AutoModel
from zrag.chunk_node import Node

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Embeddings:
    """Handles embedding generation using Hugging Face models."""

    def __init__(self, model_name: str = "nomic-ai/nomic-embed-text-v1.5", device: str = None):
        self.model_name = model_name
        self.device = device if device else ("cuda" if torch.cuda.is_available() else "cpu")
        logger.info(f"Initializing Embeddings with model: {self.model_name} on device: {self.device}")
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = AutoModel.from_pretrained(self.model_name, trust_remote_code=True).to(self.device)
        self.model.eval()

    def embed(self, texts: List[str], batch_size: int = 4) -> np.ndarray:
        """
        Generates embeddings for a list of texts.

        Args:
            texts (List[str]): List of text strings.
            batch_size (int, optional): Batch size for embedding computation. Defaults to 4.

        Returns:
            np.ndarray: A NumPy array of embeddings.
        """
        all_embeddings = []
        with torch.no_grad():
            for i in range(0, len(texts), batch_size):
                batch = texts[i:i + batch_size]
                encoded_input = self.tokenizer(batch, padding=True, truncation=True, return_tensors="pt").to(self.device)
                model_output = self.model(**encoded_input)
                batch_embeddings = self.mean_pooling(model_output, encoded_input['attention_mask']).cpu().numpy()
                all_embeddings.append(batch_embeddings)
        return np.concatenate(all_embeddings, axis=0)

    def embed_nodes(self, nodes: List[Node], batch_size: int = 4) -> List[Node]:
        """
        Generates embeddings for a list of nodes and updates the nodes in-place.

        Args:
            nodes (List[Node]): List of Node objects.
            batch_size (int, optional): Batch size for embedding computation. Defaults to 4.

        Returns:
            List[Node]: The same list of Node objects, with their 'embedding' attribute updated.
        """
        texts = [node.text for node in nodes]
        embeddings = self.embed(texts, batch_size=batch_size)  # Reusing the embed method
        for node, embedding in zip(nodes, embeddings):
            node.embedding = embedding
        return nodes

    @staticmethod
    def mean_pooling(model_output, attention_mask):
        """Applies mean pooling to the model outputs."""
        token_embeddings = model_output.last_hidden_state
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, dim=1)
        sum_mask = torch.clamp(input_mask_expanded.sum(dim=1), min=1e-9)
        return sum_embeddings / sum_mask

    @staticmethod
    def cosine_similarity(embedding1: np.ndarray, embedding2: np.ndarray) -> np.ndarray:
        """Computes cosine similarity between two embeddings or batches of embeddings."""
        embedding1_norm = np.linalg.norm(embedding1, axis=-1, keepdims=True)
        embedding2_norm = np.linalg.norm(embedding2, axis=-1, keepdims=True)

        embedding1_norm[embedding1_norm == 0] = 1e-9  # Avoid division by zero
        embedding2_norm[embedding2_norm == 0] = 1e-9

        return np.dot(embedding1, embedding2.T) / (embedding1_norm * embedding2_norm.T)


# tests/test_doc_loader.py
import unittest
import os
from pathlib import Path
from zrag.doc_loader import DocumentLoader, Document, _process_file  # Import _process_file

class TestDocumentLoader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_dir = Path("./test_data")
        cls.test_dir.mkdir(exist_ok=True)
        cls.pdf_path = cls.test_dir / "test.pdf"
        cls.md_path = cls.test_dir / "test.md"
        cls.txt_path = cls.test_dir / "test.txt"
        cls.py_path = cls.test_dir / "test.py"

        # Create dummy files
        with open(cls.pdf_path, "wb") as f:
            f.write(b"%PDF-1.5\n...")  # Minimal PDF content
        with open(cls.md_path, "w", encoding="utf-8") as f:
            f.write("# Test Markdown\nThis is a test.")
        with open(cls.txt_path, "w", encoding="utf-8") as f:
            f.write("This is a test text file.")
        with open(cls.py_path, "w", encoding="utf-8") as f:
            f.write("def test():\n    print('Hello')")

    @classmethod
    def tearDownClass(cls):
       for file in [cls.pdf_path, cls.md_path, cls.txt_path, cls.py_path]:
            if file.exists():
                os.remove(file)
       if cls.test_dir.exists():
            os.rmdir(cls.test_dir)

    def test_load_single_pdf(self):
        loader = DocumentLoader(str(self.test_dir))
        # Use _process_file directly for more control in testing:
        documents = _process_file(self.pdf_path, loader.encoding, None, None, None, code_mime_types=loader.CODE_MIME_TYPES)  # Pass code_mime_types
        self.assertEqual(len(documents), 1)
        self.assertIsInstance(documents[0], Document)
        self.assertIn("page_label", documents[0].metadata)

    def test_load_single_markdown(self):
        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(filenames=["test.md"])
        self.assertEqual(len(documents), 1)
        self.assertIn("Test Markdown", documents[0].text)

    def test_load_single_text(self):
        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(filenames=["test.txt"])
        self.assertEqual(len(documents), 1)
        self.assertIn("test text file", documents[0].text)

    def test_load_single_code(self):
        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(filenames=["test.py"])
        self.assertEqual(len(documents), 1)
        self.assertEqual(documents[0].text, "def test():\n    print('Hello')") # Check without preprocessing

    def test_load_with_preprocess(self):
        def preprocess(text):
            return text.upper()
        loader = DocumentLoader(str(self.test_dir))

        # Use _process_file directly for more control in testing:
        documents = _process_file(self.txt_path, loader.encoding, None, None, None, preprocess_fn=preprocess, code_mime_types=loader.CODE_MIME_TYPES) # Pass preprocess_fn and code_mime_types
        self.assertEqual(len(documents), 1)  # Check if a document was loaded
        self.assertEqual(documents[0].text, "THIS IS A TEST TEXT FILE.")

    def test_load_recursive(self):
        # Create a subdirectory and a file within it
        subdir = self.test_dir / "subdir"
        subdir.mkdir(exist_ok=True)
        subfile = subdir / "subtest.txt"
        with open(subfile, "w", encoding="utf-8") as f:
            f.write("This is a subdirectory test file.")

        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(recursive=True)

        # Remove the created subdirectory and file after testing
        os.remove(subfile)
        os.rmdir(subdir)
        self.assertGreaterEqual(len(documents), 4) # includes files from the recursive search

    def test_load_with_ext_filter(self):
        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(ext=["*.md", "*.txt"])
        self.assertEqual(len(documents), 2)

    def test_load_with_exc_filter(self):
        loader = DocumentLoader(str(self.test_dir))
        documents = loader.load(exc=["*.pdf"])
        self.assertEqual(len(documents), 3)




if __name__ == "__main__":
    unittest.main()


import os
import uuid
import fitz
import logging
import datetime
import markdown
import mimetypes
from pathlib import Path
from typing import List, Dict, Optional, Callable
from concurrent.futures import ProcessPoolExecutor, as_completed

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Document:
    """Represents a single document with its content and metadata."""

    def __init__(self, document_id: str, metadata: Dict, text: str):
        self.document_id = document_id
        self.metadata = metadata
        self.text = text

    def __repr__(self):
        return f"Document(document_id='{self.document_id}', metadata={self.metadata}, text='{self.text[:20]}...')"


def _process_file(
    file_path: Path,
    encoding: str,
    ext: Optional[List[str]],
    exc: Optional[List[str]],
    filenames: Optional[List[str]],
    preprocess_fn: Optional[Callable[[str], str]] = None,
    code_mime_types=None  # Added to receive CODE_MIME_TYPES
) -> List[Document]:
    """
    Helper function to process a single file based on filtering criteria.
    """
    from zrag.doc_loader import DocumentLoader, Document  # Import inside the function

    # Check if the file should be processed based on filename
    if filenames is not None and file_path.name not in filenames:
        return []

    if ext is not None:
        if not any(file_path.match(pattern) for pattern in ext):
            return []

    if exc is not None:
        if any(file_path.match(pattern) for pattern in exc):
            return []

    mime_type, _ = mimetypes.guess_type(str(file_path))

    # Check for code MIME types using the passed dictionary
    if code_mime_types is not None and mime_type in code_mime_types:
        return DocumentLoader._read_code(file_path, encoding, preprocess_fn)

    try:
        return DocumentLoader._read_file(file_path, encoding, preprocess_fn)
    except Exception as e:
        logger.error(f"Error reading file {file_path}: {e}")
        return []


class DocumentLoader:
    """
    Loads and reads files from a directory, returning documents with
    metadata and text content.
    """

    CODE_MIME_TYPES = {
        'text/x-python': '.py',
        'text/javascript': '.js',
        'text/x-java-source': '.java',
        'text/x-c++src': '.cpp',
        'text/x-csrc': '.c',
        'text/x-ruby': '.rb',
        'text/x-go': '.go',
        'text/x-shellscript': '.sh',
        'application/typescript': '.ts',
        # Add more as needed
    }

    def __init__(self, directory_path: str, encoding: str = "utf-8"):
        self.directory_path = directory_path
        self.encoding = encoding
        self._initialize_mimetypes()

    def _initialize_mimetypes(self):
        """Ensures code MIME types are recognized."""
        for mime, ext in self.CODE_MIME_TYPES.items():
            mimetypes.add_type(mime, ext)

    @staticmethod
    def _read_file(
        file_path: Path,
        encoding: str,
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """Reads a file based on its MIME type."""
        mime_type, _ = mimetypes.guess_type(str(file_path))

        read_methods = {
            "application/pdf": DocumentLoader._read_pdf,
            "text/markdown": DocumentLoader._read_markdown,
            "text/plain": DocumentLoader._read_text,
            # code MIME types
            "text/x-python": DocumentLoader._read_code,
            "text/javascript": DocumentLoader._read_code,
            "application/typescript": DocumentLoader._read_code,
            "text/x-java-source": DocumentLoader._read_code,
            "text/x-c++src": DocumentLoader._read_code,
            "text/x-csrc": DocumentLoader._read_code,
            "text/x-ruby": DocumentLoader._read_code,
            "text/x-go": DocumentLoader._read_code,
            "text/x-shellscript": DocumentLoader._read_code,
        }

        read_method = read_methods.get(mime_type, DocumentLoader._read_text)
        return read_method(file_path, encoding, preprocess_fn)

    @staticmethod
    def _read_pdf(
        file_path: Path,
        encoding: str,
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """Reads a PDF file."""
        documents = []
        try:
            doc = fitz.open(str(file_path))
        except Exception as e:
            logger.error(f"Failed to open PDF file {file_path}: {e}")
            return documents

        file_stats = file_path.stat()
        metadata_common = {
            'file_name': file_path.name,
            'file_path': str(file_path),
            'file_type': 'application/pdf',
            'file_size': file_stats.st_size,
            'creation_date': datetime.datetime.fromtimestamp(file_stats.st_ctime).strftime('%Y-%m-%d'),
            'last_modified_date': datetime.datetime.fromtimestamp(file_stats.st_mtime).strftime('%Y-%m-%d')
        }

        for page_num, page in enumerate(doc):
            try:
                text = page.get_text()
                if preprocess_fn:
                    text = preprocess_fn(text)
            except Exception as e:
                logger.warning(f"Failed to extract or preprocess text from page {page_num+1} in {file_path}: {e}")
                text = ""

            metadata = metadata_common.copy()
            metadata.update({
                'page_label': str(page_num + 1)
            })

            document_id = str(uuid.uuid4())
            documents.append(Document(document_id, metadata, text))

        doc.close()
        return documents

    @staticmethod
    def _read_markdown(
        file_path: Path,
        encoding: str,
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """Reads a Markdown file."""
        try:
            with open(file_path, encoding=encoding) as f:
                md_content = f.read()
            # Convert Markdown to plain text
            text = markdown.markdown(md_content)
            if preprocess_fn:
                text = preprocess_fn(text)
        except Exception as e:
            logger.error(f"Failed to read or preprocess Markdown file {file_path}: {e}")
            text = ""

        return [DocumentLoader._create_single_document(file_path, text, 'text/markdown')]

    @staticmethod
    def _read_text(file_path: Path, encoding: str, preprocess_fn: Optional[Callable[[str], str]] = None) -> List[Document]:
        """Reads a plain text file."""
        try:
            with open(file_path, encoding=encoding) as f:
                text = f.read()
            if preprocess_fn:
                text = preprocess_fn(text)
        except Exception as e:
            logger.error(f"Failed to read or preprocess text file {file_path}: {e}")
            text = ""

        return [DocumentLoader._create_single_document(file_path, text, 'text/plain')]


    @staticmethod
    def _read_code(
        file_path: Path,
        encoding: str,
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """Reads a code file."""
        try:
            with open(file_path, encoding=encoding) as f:
                code = f.read()
            if preprocess_fn:
                code = preprocess_fn(code)
        except Exception as e:
            logger.error(f"Failed to read code file {file_path}: {e}")
            code = ""

        mime_type = mimetypes.guess_type(str(file_path))[0] or 'text/plain'
        return [DocumentLoader._create_single_document(file_path, code, mime_type)]

    @staticmethod
    def _create_single_document(file_path: Path, text: str, mime_type: str) -> Document:
        """Creates a Document object with metadata."""
        file_stats = file_path.stat()
        metadata = {
            'page_label': '1',
            'file_name': file_path.name,
            'file_path': str(file_path),
            'file_type': mime_type,
            'file_size': file_stats.st_size,
            'creation_date': datetime.datetime.fromtimestamp(file_stats.st_ctime).strftime('%Y-%m-%d'),
            'last_modified_date': datetime.datetime.fromtimestamp(file_stats.st_mtime).strftime('%Y-%m-%d')
        }
        document_id = str(uuid.uuid4())
        return Document(document_id, metadata, text)

    @staticmethod
    def _process_file(
        file_path: Path,
        encoding: str,
        ext: Optional[List[str]],
        exc: Optional[List[str]],
        filenames: Optional[List[str]],
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """
        Helper function to process a single file based on filtering criteria.
        """
        # Check if the file should be processed based on filename
        if filenames is not None and file_path.name not in filenames:
            return []

        if ext is not None:
            if not any(file_path.match(pattern) for pattern in ext):
                return []

        if exc is not None:
            if any(file_path.match(pattern) for pattern in exc):
                return []

        try:
            return DocumentLoader._read_file(file_path, encoding, preprocess_fn)
        except Exception as e:
            logger.error(f"Error reading file {file_path}: {e}")
            return []

    def load(
        self,
        recursive: bool = False,
        ext: Optional[List[str]] = None,
        exc: Optional[List[str]] = None,
        filenames: Optional[List[str]] = None,
        max_workers: int = os.cpu_count(),
        preprocess_fn: Optional[Callable[[str], str]] = None
    ) -> List[Document]:
        """
        Loads documents from the directory, applying optional filtering and preprocessing.
        """
        directory = Path(self.directory_path)
        documents: List[Document] = []

        file_generator = directory.rglob("*") if recursive else directory.glob("*")
        file_paths = (fp for fp in file_generator if fp.is_file())

        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            futures = {
                executor.submit(
                    _process_file,  # Call the external function
                    file_path,
                    self.encoding,
                    ext,
                    exc,
                    filenames,
                    preprocess_fn,
                    self.CODE_MIME_TYPES  # Pass CODE_MIME_TYPES
                ): file_path for file_path in file_paths
            }

            for future in as_completed(futures):
                try:
                    result = future.result()
                    documents.extend(result)
                except Exception as e:
                    logger.error(f"Error processing file {futures[future]}: {e}")

        return documents


import logging
import json
import csv
import os
from typing import List, Dict, Any, Optional, Union, Iterator
from zrag.doc_loader import DocumentLoader
from zrag.llm import LLM
from zrag.prompt_manager import PromptManager
from zrag.chunk_node import get_chunk_splitter, ChunkSplitter, Node
from zrag.embeddings import Embeddings

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class DataGenerator:  
    """
    Generates datasets using LLMs, knowledge sources, and example datasets.
    """
    def __init__(
        self,
        file_loader: DocumentLoader,  
        chunk_splitter: ChunkSplitter,  
        embeddings: Embeddings,  
        llm: LLM, 
        prompt_manager: PromptManager,  
        example_dataset_path: str,  
        output_format: str = "json",
        output_path: str = "generated_dataset.json",
        batch_size: int = 8,  
        default_prompt_template: str = "dataset_instruction" 
    ):
        """
        Initializes the DataGenerator with necessary components.
        """
        self.file_loader = file_loader
        self.chunk_splitter = chunk_splitter
        self.embeddings = embeddings
        self.llm = llm
        self.prompt_manager = prompt_manager

        self.example_dataset_path = example_dataset_path
        self.output_format = output_format.lower()
        self.output_path = output_path
        self.batch_size = batch_size
        self.default_prompt_template = default_prompt_template

    def load_knowledge(self, directory_path: str, **kwargs) -> List[Node]:  
        """Loads, chunks, and embeds knowledge documents."""

        documents = self.file_loader.load(directory_path=directory_path, **kwargs)  
        chunks = self.chunk_splitter.split(documents)
        self.embeddings.embed_nodes(chunks) 
        return chunks 

    def _load_example_dataset(self) -> List[Dict[str, Any]]:  
        """Loads the example dataset."""
        if not os.path.exists(self.example_dataset_path): 
            raise FileNotFoundError(f"Example dataset file not found: {self.example_dataset_path}")

        try:
            if self.example_dataset_path.endswith(".json"):
                with open(self.example_dataset_path, "r", encoding="utf-8") as f:
                    return json.load(f)
            elif self.example_dataset_path.endswith(".csv"):
                with open(self.example_dataset_path, "r", encoding="utf-8") as f:
                    reader = csv.DictReader(f)
                    return list(reader)
            else:
                raise ValueError("Unsupported example dataset format. Use JSON or CSV.")  
        except Exception as e:  
            raise RuntimeError(f"Error loading example dataset: {e}") from e  

    def generate_dataset(
        self,
        knowledge_nodes: List[Node],
        num_entries: int = 1000,
        prompt_template: Optional[str] = None,
        **llm_kwargs
    ):
        """Generates a dataset using knowledge nodes and the example dataset format."""

        examples = self._load_example_dataset() 

        if not examples:
            raise ValueError("Example dataset is empty.")  

        # Default prompt_template handling
        if not prompt_template:
            prompt_template = self.default_prompt_template

        # Generator to produce prompts in a streaming way
        def prompt_generator(num: int, knowledge: List[Node], template_name: str) -> Iterator[str]:
            """Generate prompts from knowledge nodes in a streaming way"""
            for _ in range(num):
                knowledge_node = knowledge[_ % len(knowledge)] # reusing knowledge nodes 
                variables = {"instruction": "Generate a question and answer pair based on the following context.",
                               "input_data": knowledge_node.text}
                prompt = self.prompt_manager.create_prompt(template_name=template_name, **variables)
                if prompt:
                    yield prompt
                else:
                    logger.warning("Failed to generate prompt for a dataset entry.")
                    continue

        # Generate dataset entries using LLM (streaming)
        generated_entries = self.llm.generate(list(prompt_generator(num_entries, knowledge_nodes, prompt_template)), **llm_kwargs)  

        # Function to structure generated entries
        def structure_entry(entry: str):
            parts = entry.split('\n')  # Split by newline 
            if len(parts) >= 2:
                return {"question": parts[0].strip(), "answer": " ".join(parts[1:]).strip()}  
            else:
                logger.warning("Generated entry does not conform to expected format. Skipping.")
                return None  

        # Structure generated entries 
        structured_entries = [
            structure_entry(entry) for entry in generated_entries if structure_entry(entry)
        ]

        self._save_dataset(structured_entries)  

    def _save_dataset(self, dataset: List[Dict[str, Any]]):  
        """Saves the generated dataset."""

        if not dataset:
            logger.warning("Dataset is empty. Nothing to save.")
            return

        try:
            if self.output_format == "json":
                with open(self.output_path, "w", encoding="utf-8") as f:  
                    json.dump(dataset, f, ensure_ascii=False, indent=4)
            elif self.output_format == "csv":
                fieldnames = dataset[0].keys()  
                with open(self.output_path, "w", newline="", encoding="utf-8") as f: 
                    writer = csv.DictWriter(f, fieldnames=fieldnames)
                    writer.writeheader()
                    writer.writerows(dataset)
            else:
                raise ValueError("Unsupported output format. Use 'json' or 'csv'.")  
        except Exception as e:  
            raise RuntimeError(f"Error saving dataset: {e}") from e 


import unittest
import numpy as np
from unittest.mock import patch, MagicMock

from zrag.embeddings import Embeddings
from zrag.chunk_node import Node


class TestEmbeddings(unittest.TestCase):
    """Comprehensive tests for the Embeddings class."""

    @patch("zrag.embeddings.AutoModel.from_pretrained")
    @patch("zrag.embeddings.AutoTokenizer.from_pretrained")
    def setUp(self, mock_tokenizer, mock_model):
        """Sets up mocks for the Hugging Face model and tokenizer."""
        self.mock_model = mock_model.return_value
        self.mock_tokenizer = mock_tokenizer.return_value
        self.embeddings = Embeddings()  # Using default model name

    def test_initialization(self):
        """Tests initialization of the Embeddings class with default values."""
        self.assertEqual(self.embeddings.model_name, "nomic-ai/nomic-embed-text-v1.5")
        self.assertIn(self.embeddings.device, ["cuda", "cpu"])

    def test_embed_single_text(self):
        """Tests embedding a single text string."""
        text = "Hello world"
        self.mock_tokenizer.__call__.return_value = {
            "input_ids": MagicMock(),
            "attention_mask": MagicMock(),
        }
        self.mock_model.__call__.return_value = MagicMock(
            last_hidden_state=MagicMock()
        )

        with patch.object(
            Embeddings,
            "mean_pooling",
            return_value=MagicMock(
                cpu=lambda: MagicMock(numpy=lambda: np.array([0.1, 0.2]))
            ),
        ):
            embedding = self.embeddings.embed([text])  # Pass as a list

        self.assertIsInstance(embedding, np.ndarray)
        self.assertEqual(embedding.shape, (1, 2))  # 1 text, 2-dimensional embedding

    def test_embed_multiple_texts(self):
        """Tests embedding multiple text strings."""
        texts = ["Hello world", "Test embedding"]
        self.mock_tokenizer.__call__.return_value = {
            "input_ids": MagicMock(),
            "attention_mask": MagicMock(),
        }
        self.mock_model.__call__.return_value = MagicMock(
            last_hidden_state=MagicMock()
        )

        with patch.object(
            Embeddings,
            "mean_pooling",
            return_value=MagicMock(
                cpu=lambda: MagicMock(numpy=lambda: np.array([[0.1, 0.2], [0.3, 0.4]]))
            ),
        ):
            embeddings = self.embeddings.embed(texts)

        self.assertIsInstance(embeddings, np.ndarray)
        self.assertEqual(embeddings.shape, (2, 2))  # 2 texts, 2-dimensional embeddings

    def test_embed_nodes(self):
        """Tests embedding a list of Node objects."""
        nodes = [Node("Hello world", {}), Node("Test embedding", {})]
        with patch.object(
            Embeddings, "embed", return_value=np.array([[0.1, 0.2], [0.3, 0.4]])
        ):
            result_nodes = self.embeddings.embed_nodes(nodes)

        self.assertEqual(len(result_nodes), 2)
        self.assertIsInstance(result_nodes[0].embedding, np.ndarray)
        self.assertEqual(result_nodes[0].embedding.shape, (2,))
        np.testing.assert_array_equal(result_nodes[0].embedding, [0.1, 0.2])
        np.testing.assert_array_equal(result_nodes[1].embedding, [0.3, 0.4])

    def test_cosine_similarity_single(self):
        """Tests cosine similarity between single embeddings."""
        embedding1 = np.array([1, 0])
        embedding2 = np.array([0, 1])
        similarity = Embeddings.cosine_similarity(embedding1, embedding2)
        self.assertAlmostEqual(similarity[0][0], 0.0)

        embedding3 = np.array([1, 1])
        similarity = Embeddings.cosine_similarity(embedding1, embedding3)
        self.assertAlmostEqual(similarity[0][0], 1.0 / np.sqrt(2))

    def test_cosine_similarity_batch(self):
        """Tests cosine similarity between batches of embeddings."""
        embeddings1 = np.array([[1, 0], [0, 1]])
        embeddings2 = np.array([[0, 1], [1, 0]])
        similarity_matrix = Embeddings.cosine_similarity(embeddings1, embeddings2)

        # Check the shape and values of the similarity matrix
        self.assertEqual(similarity_matrix.shape, (2, 2))
        # Add assertions to verify the expected similarity values

    def test_empty_text_list(self):
        """Tests embedding an empty list of texts."""
        with patch.object(
            Embeddings,
            "mean_pooling",
            return_value=MagicMock(
                cpu=lambda: MagicMock(numpy=lambda: np.array([]))
            ),
        ):
            embeddings = self.embeddings.embed([])  # Pass an empty list

        self.assertIsInstance(embeddings, np.ndarray)
        self.assertEqual(embeddings.shape, (0,))  # Expect an empty array


if __name__ == "__main__":
    unittest.main()